from Crypto.Util.number import inverse
from sqlalchemy.orm.exc import NoResultFound

from belenios.custom_types.belenios.EventType import EventType
from belenios.dataclass.dataclass import EncryptedText
from belenios.models.belenios.CiphertextModel import CiphertextModel
from belenios.models.belenios.CredentialModel import CredentialModel
from belenios.models.belenios.EncryptedTallyRowModel import EncryptedTallyRowModel
from belenios.models.belenios.PedersenTrusteeModel import PedersenTrusteeModel
from belenios.models.belenios.ResultModel import ResultModel
from belenios.models.belenios.ResultRowModel import ResultRowModel
from belenios.models.belenios.SingleTrusteeModel import SingleTrusteeModel
from belenios.models.belenios.TallyModel import TallyModel
from belenios.models.belenios.VoterModel import VoterModel
from belenios.utilities.ElGamal import ElGamal
from belenios.utilities.TrusteesKeysProtocol import TrusteesKeysProtocol
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.DbSession import DbSession


class ServerCore:
    @staticmethod
    def make_election_unique(election):
        election.uuid = Utility.generate_base58_string(14)
        DbSession().session.commit()

    @staticmethod
    def check_voters_credentials_and_salts(election_uuid, public_credentials, salts):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return False
        election = election_bundle.election
        if election is None:
            return False
        salts_seen = []
        public_credential_seen = []
        for index, pc in public_credentials.items():
            try:
                existing_voter = DbSession().session.query(VoterModel).filter_by(email=pc["email"]).first()
            except NoResultFound:
                return False
            if str(pc["index"]) != str(index):
                return False
            if pc["weight"] != existing_voter.weight:
                return False
            if pc["public_credential"] in public_credential_seen:
                return False
            if salts[pc["index"]] in salts_seen:
                return False
            salts_seen.append(salts[pc["index"]])
            public_credential_seen.append(pc["public_credential"])

        if len(public_credential_seen) != len(election_bundle.voters) or len(salts_seen) != len(election_bundle.voters):
            return False
        return True

    @staticmethod
    def import_public_key_single_trustee(election_uuid, trustee_public_key, trustee_id):
        single_trustee = SingleTrusteeModel()
        single_trustee.trustee_id = trustee_id
        single_trustee.trustee_public_key = trustee_public_key
        print(trustee_public_key)
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        # ServerAdministratorCore().add_single_trustees(self.app.pargs.uuid, int(self.app.pargs.count))

        # 2. S checks γ
        # We can check is the key and the pok is correctly generated :
        if TrusteesKeysProtocol().check_proof_trustee_public_key(election_bundle.election,
                                                                 single_trustee.trustee_public_key):
            print("Trustee key is valid")
        else:
            print("Trustee key is invalid")
            return False

        election_bundle.trustee_kinds.append(single_trustee)
        DbSession().session.commit()
        return single_trustee

    @staticmethod
    def import_cert_pedersen_trustee(election_uuid, signed_msg, trustee_id):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            pedersen_trustee = PedersenTrusteeModel()
            pedersen_trustee.threshold = 2
            election_bundle.trustee_kinds.append(pedersen_trustee)
        pedersen_trustee.append_trustee_id(trustee_id)

        # ServerAdministratorCore().add_single_trustees(self.app.pargs.uuid, int(self.app.pargs.count))

        # 2. S checks γ
        # We can check is the cert and the signed_msg is correctly generated :
        if Utility.check_signed_msg(signed_msg, signed_msg.signed_object.verification, election_bundle.election):
            print("Trustee cert is valid")
        else:
            print("Trustee cert is invalid")
            return False

        pedersen_trustee.certs.append(signed_msg)
        # print(pedersen_trustee)
        # print(trustee_kind_row)
        # print(trustee_kind_row.trustee_kinds)
        DbSession().session.commit()
        return pedersen_trustee

    @staticmethod
    def import_coefexps_verification_keys_pedersen_trustee(election_uuid, v_output):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        pedersen_trustee = None
        for pedersen_trustee_iter in election_bundle.trustee_kinds:
            if pedersen_trustee_iter.trustee_kind_type == "PedersenTrusteeModel":
                pedersen_trustee = pedersen_trustee_iter
                break
        if pedersen_trustee is None:
            pedersen_trustee = PedersenTrusteeModel()
            pedersen_trustee.threshold = 2
            election_bundle.trustee_kinds.append(pedersen_trustee)

        if len(election_bundle.polynomials) > len(pedersen_trustee.coefexps):
            pedersen_trustee.coefexps = []
            for p_i in election_bundle.polynomials:
                pedersen_trustee.coefexps.append(p_i.coefexps)
        pedersen_trustee.verification_keys.append(v_output.trustee_public_key)
        DbSession().session.commit()
        return pedersen_trustee

    @staticmethod
    def compute_election_public_key(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        # 10. S creates the election E
        election_bundle.election.public_key = TrusteesKeysProtocol().compute_election_public_key(election_bundle)
        print("Election public key is %s" % election_bundle.election.public_key)
        DbSession().session.commit()

        return election_bundle

    @staticmethod
    def add_questions(election_uuid, questions):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        for question in questions:
            election_bundle.election.questions.append(question)
        DbSession().session.commit()

        return questions

    @staticmethod
    def make_election_immutable(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        # Fix ElectionModel, trustee_kinds, public_credentials and publish the bundle to avoid later manipulation and
        election_bundle.setup_data.election = election_bundle.election.to_hash_sha256()
        print("Election hash is %s" % election_bundle.setup_data.election)

        DbSession().session.commit()
        return election_bundle

    @staticmethod
    def print_election_bundle(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        print(election_bundle)

    @staticmethod
    def add_to_tally(election_uuid, ballot_to_add):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        tally = DbSession().session.query(TallyModel).filter_by(election_bundle=election_bundle).first()
        if tally is None:
            print("Tally invalid")
            return None

        replaced = False
        for index, ballot in enumerate(tally.ballots):
            if ballot.credential == ballot_to_add.credential:
                tally.ballots[index] = ballot_to_add
                replaced = True
                break
        if not replaced:
            tally.ballots.append(ballot_to_add)
        DbSession().session.commit()
        return tally

    @staticmethod
    def compute_encrypted_tally(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        election = election_bundle.election
        tally = DbSession().session.query(TallyModel).filter_by(election_bundle=election_bundle).first()
        if tally is None:
            print("Tally invalid")
            return None

        for index_question, question in enumerate(election.questions):
            encrypted_tally_row = EncryptedTallyRowModel()
            is_blank = question.type == "QuestionHModel" and question.blank
            for index_answer in range(len(question.answers) + int(is_blank)):  # loop over choices + blank
                if question.type == 'QuestionHModel':
                    # Initialize the product of ciphertexts for this answer
                    product_ciphertext = [1, 1]

                    # Iterate over each ballot
                    for ballot in election_bundle.tally.ballots:
                        answer = ballot.answers[index_question]
                        choice = answer.choices[index_answer]
                        credential = DbSession().session.query(CredentialModel).filter_by(
                            public_credential=ballot.credential).first()
                        if credential is None:
                            print("Credential not found.")
                            return None
                        weight = credential.weight
                        # Multiply the corresponding ciphertext to the power of weight
                        product_ciphertext[0] *= (choice.alpha ** weight)
                        product_ciphertext[1] *= (choice.beta ** weight)
                        product_ciphertext[0] %= election.group.p
                        product_ciphertext[1] %= election.group.p
                        # ciphertext_answer = answer[index]
                        # product_ciphertext = (
                        #     product_ciphertext[0] * ciphertext_answer[0],
                        #     product_ciphertext[1] * ciphertext_answer[1])

                    # Append the resulting ciphertext to the array
                    ciphertext = CiphertextModel()
                    ciphertext.alpha = product_ciphertext[0]
                    ciphertext.beta = product_ciphertext[1]
                    encrypted_tally_row.ciphertexts.append(ciphertext)
                else:
                    print("not implemented")
                    return None
                    # # Initialize the array of ciphertexts for this answer
                    # answer_ciphertexts = []
                    #
                    # # Iterate over each ballot
                    # for ballot in ballots:
                    #     answer = ballot.answers[index]
                    #     weight = ballot.weight
                    #
                    #     # Ensure weight is 1 for non-homomorphic questions
                    #     if weight != 1:
                    #         raise ValueError("Non-homomorphic question has weight other than 1")
                    #
                    #     # Append the corresponding ciphertext
                    #     ciphertext_answer = answer[index]
                    #     answer_ciphertexts.append(ciphertext_answer)
                    #
                    # # Append the array of ciphertexts to the array
                    # ciphertexts_question.append(answer_ciphertexts)

            # Append the array of ciphertexts for this question to the encrypted tally array
            election_bundle.tally.encrypted_tally_rows.append(encrypted_tally_row)
        tally.sized_encrypted_tally.num_tallied = len(election_bundle.tally.ballots)

        tally.sized_encrypted_tally.total_weight = 0
        for ballot in election_bundle.tally.ballots:
            credential = DbSession().session.query(CredentialModel).filter_by(
                public_credential=ballot.credential).first()
            if credential is None:
                print("Credential not found.")
                return None
            tally.sized_encrypted_tally.total_weight += credential.weight
        # TODO: hash the tally
        # tally.sized_encrypted_tally.encrypted_tally  =
        DbSession().session.commit()
        return tally

    @staticmethod
    def set_event_encrypted_tally(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        event = Utility.put_new_event(election_bundle)
        event.type = EventType.EncryptedTally
        DbSession().session.commit()
        return event

    @staticmethod
    def add_event_partial_decryption(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        event = Utility.put_new_event(election_bundle)
        event.type = EventType.PartialDecryption
        DbSession().session.commit()
        return event

    @staticmethod
    def lambda_factor(trustee_ids, delta, group):
        l = 1
        for trustee_id in trustee_ids:
            if trustee_id == delta:
                continue
            # print("j", inverse(trustee_id - delta, group.q))
            l *= trustee_id * inverse(trustee_id - delta, group.q)
        return l % group.q

    @staticmethod
    def compute_result(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        trustees_number = 0
        for trustee_kind in election_bundle.trustee_kinds:
            if trustee_kind.trustee_kind_type == "SingleTrusteeModel":
                trustees_number += 1
            else:
                trustees_number += len(trustee_kind.trustee_ids)
                # TODO: handle correctly group of trustees

        if len(election_bundle.owned_partial_decryptions) < trustees_number:
            print("Not enough partial decryption")
            return None
        group = election_bundle.election.group
        # g = election_bundle.election.group.g
        # q = election_bundle.election.group.q
        # total_weight = election_bundle.tally.sized_encrypted_tally.total_weight
        result = ResultModel()
        for i, question in enumerate(election_bundle.election.questions):
            result_row = ResultRowModel()
            is_blank = question.type == "QuestionHModel" and question.blank
            for j in range(len(question.answers) + int(is_blank)):
                F_ij = 1
                for owned_partial_decryption in election_bundle.owned_partial_decryptions:
                    trustee_kind = Utility.get_trustee_kind_from_id(owned_partial_decryption.trustee_id,
                                                                    election_bundle)
                    partial_decryption = owned_partial_decryption.partial_decryption
                    if trustee_kind.trustee_kind_type == "SingleTrusteeModel":
                        F_ijt = partial_decryption.decryption_factor_rows[i].decryption_factors[j]
                    else:
                        lambda_delta_I = ServerCore.lambda_factor(trustee_kind.trustee_ids,
                                                                  owned_partial_decryption.trustee_id, group)
                        F_ijt = pow(partial_decryption.decryption_factor_rows[i].decryption_factors[j], lambda_delta_I,
                                    group.p)
                    F_ij *= F_ijt

                aij = election_bundle.tally.encrypted_tally_rows[i].ciphertexts[j]
                alpha_aij = aij.alpha
                beta_aij = aij.beta
                a = EncryptedText([(alpha_aij, beta_aij)])

                res_ij = ElGamal().decrypt_from_int(a, private_key=None, group=group, exponential=True, right=F_ij,
                                                    max_dlp=10 ** 5)

                # m_prime = (beta_aij * inverse(F_ij, q)) % q
                # print("jjj", beta_aij * F_ij)
                # teta = beta_aij / F_ij
                # teta = (beta_aij * inverse(F_ij, q)) % q
                # res_ij = ElGamal().discrete_logarithm(teta, g, total_weight*10)
                # print("res_ij",res_ij)
                # res_ij = ElGamal().discrete_logarithm(teta, g, q)
                result_row.result_cells.append(res_ij)
            result.result_rows.append(result_row)
        election_bundle.result = result
        DbSession().session.commit()
        return result

    @staticmethod
    def set_event_result(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        event = Utility.put_new_event(election_bundle)
        event.type = EventType.Result
        DbSession().session.commit()
        return event

    @staticmethod
    def get_array_of_result(election_uuid):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        result = []

        for i, result_row in enumerate(election_bundle.result.result_rows):
            row = []
            for j, result_cell in enumerate(result_row.result_cells):
                row.append(result_cell)
            result.append(row)

        return result



from sqlalchemy.types import TypeDecorator, VARCHAR


class UuidType(TypeDecorator):
    impl = VARCHAR
    cache_ok = True
    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if len(value) < 14:
            raise ValueError("Value must be at least 14 characters long")
        for letter in value:
            if letter not in "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz":
                raise ValueError("Value must be base58 encoded")
        return value

    def process_result_value(self, value, dialect):
        return value

    def process_literal_param(self, value, dialect):
        return value



from setuptools import setup, find_packages

# from belenios.core.version import get_version

# VERSION = get_version()

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='belenios',
    version='0.0.4',
    description='This is a proof of concept of the Belenios election protocol.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='BISCHOFF Taylor, BRIONGOS Alexandre, ILLA Mohamed abdallahi, MALIGUE Dylan',
    author_email='alexandre.briongos@etu.univ-amu.fr',
    url='https://etulab.univ-amu.fr/b21211803/projet-ter-2024-belenios',
    license='unlicensed',
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    package_data={'belenios': ['templates/*']},
    include_package_data=True,
    entry_points="""
        [console_scripts]
        belenios = belenios.main:main
    """,
    install_requires=[
        'cement[colorlog,jinja2,yaml]==3.0.8',
        'SQLAlchemy==2.0.25',
        'pycryptodome==3.20.0',
        'base58==2.1.1',
    ],
)



from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.ArrayOfGsType import ArrayOfGsType
from belenios.models.BaseModel import BaseModel


class DecryptionFactorRowModel(BaseModel):
    partial_decryption_id = Column(Integer, ForeignKey('partial_decryption_model.id'))
    decryption_factors = Column(ArrayOfGsType, nullable=False)

    def __init__(self, *args, **kwargs):
        if 'decryption_factors' not in kwargs:
            kwargs['decryption_factors'] = []
        super().__init__(*args, **kwargs)


from sqlalchemy import Column, Integer, ForeignKey

from belenios.custom_types.belenios.ArrayOfGsType import ArrayOfGsType
from belenios.models.belenios.SignedObjectModel import SignedObjectModel


class RawCoefexpsModel(SignedObjectModel):
    id = Column(Integer, ForeignKey('signed_object_model.id'), primary_key=True)
    raw_coefexps = Column(ArrayOfGsType, nullable=False)
    __mapper_args__ = {
        "polymorphic_identity": "RawCoefexpsModel",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'RawCoefexpsModel'

    def __str__(self):
        return (
            f"RawCoefexpsModel(\n"
            f"    raw_coefexps={self.raw_coefexps}\n"
            f")"
        )
    def to_json(self):
        return {
            "raw_coefexps": self.raw_coefexps,
        }

    @staticmethod
    def load_from_json(dict):
        return RawCoefexpsModel(type=dict['type'], raw_coefexps=dict['raw_coefexps'])



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, Mapped

from belenios.custom_types.belenios.IType import IType
from belenios.models.belenios.TrusteeKindModel import TrusteeKindModel


class SingleTrusteeModel(TrusteeKindModel):
    id = Column(Integer, ForeignKey('trustee_kind_model.id'), primary_key=True)
    trustee_id = Column(IType, nullable=False)
    trustee_public_key_id: int = Column(Integer, ForeignKey('trustee_public_key_model.id'), nullable=False)
    trustee_public_key: Mapped['TrusteePublicKeyModel'] = relationship('TrusteePublicKeyModel')
    __mapper_args__ = {
        "polymorphic_identity": "SingleTrusteeModel"
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trustee_kind_type = 'SingleTrusteeModel'



import json

from cement import Controller, ex
from cement.utils import shell

from belenios.core.belenios.ServerAdministratorCore import ServerAdministratorCore
from belenios.core.belenios.ServerCore import ServerCore
from belenios.core.belenios.TrusteeCore import TrusteeCore
from belenios.core.belenios.VoterCore import VoterCore


class ElectionCmd(Controller):
    class Meta:
        label = 'election'
        stacked_on = 'base'
        stacked_type = 'nested'
        # base_controller = True

        # text displayed at the top of --help output
        description = 'Belenios command-line tool to create and setup an election.'

        # text displayed at the bottom of --help output
        epilog = 'Usage: belenios election'

        arguments = [
            # (['-start', '--start'],
            #  {'action': 'start',
            #   '': }),
        ]


    @ex(
        help='Print the election bundle',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
        ],
    )
    def print_election(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        ServerCore().print_election_bundle(self.app.pargs.uuid)


    @ex(
        help='Generate ballot of the voter for the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-c', '--choices'], {'help': 'The choices of the voter for the election', 'action': 'store', 'dest': 'choices'}),
            (['-p', '--privcred'], {'help': 'The private credential of the voter for the election', 'action': 'store', 'dest': 'privcred'}),
        ],
    )
    def generate_ballot(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.choices:
            self.app.pargs.choices = shell.Prompt("Enter the choices of the voter for the election: ").input

        if not self.app.pargs.privcred:
            self.app.pargs.privcred = shell.Prompt("Enter the private credential of the voter for the election: ").input

        choices = json.loads(self.app.pargs.choices)
        if not VoterCore().check_choices_structure(self.app.pargs.uuid,choices ):
            print("Choice invalid")
            return False

        ballot = VoterCore().generate_ballot(self.app.pargs.uuid, choices, self.app.pargs.privcred)
        if ballot is None:
            print("Ballot not created")
            return False
        tally = ServerCore.add_to_tally(self.app.pargs.uuid, ballot)
        if tally is None:
            self.print("Tally invalid")
            return False
        print("Ballot created.")
        return True

    @ex(
        help='Close the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
        ],
    )
    def end_ballot(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input
        event = ServerAdministratorCore.end_ballot(self.app.pargs.uuid)
        if event is None:
            print("Election not closed.")
        print("Election closed.")
        return True

    @ex(
        help='Compute the encrypted tally of the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
        ],
    )
    def compute_encrypted_tally(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        tally = ServerAdministratorCore.make_server_make_encrypted_tally(self.app.pargs.uuid)
        if tally is None:
            print("Encrypted tally not computed.")
            return False
        event = ServerCore.set_event_encrypted_tally(self.app.pargs.uuid)
        if event is None:
            print("Encrypted tally not computed.")
            return False
        print("Encrypted tally computed.")
        return True

    @ex(
        help='Decrypt partially the encrypted tally of the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fprk', '--file-private-key'],
             {'help': 'The file where to input private key of the trustee', 'action': 'store',
              'dest': 'file_private_key'}),
            (['-t', '--trustee-id'],
             {'help': 'The trustee id for the election', 'action': 'store', 'dest': 'trustee_id'}),
        ],
    )
    def decrypt(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_private_key:
            self.app.pargs.file_private_key = shell.Prompt(
                "Enter the file where to input private key of the trustee: ").input

        if not self.app.pargs.trustee_id:
            self.app.pargs.trustee_id = shell.Prompt("Enter the trustee id for the election: ").input

        try:
            with open(self.app.pargs.file_private_key, 'r') as file:
                data = json.load(file)
        except:
            print("Invalid private key file")
            return False

        owned_partial_decryption = TrusteeCore.partial_decrypt_encrypted_ballot(self.app.pargs.uuid, int(data),
                                                                                int(self.app.pargs.trustee_id))
        if owned_partial_decryption is None:
            print("Encrypted tally not partially decrypted.")
            return False
        event = ServerCore.add_event_partial_decryption(self.app.pargs.uuid)
        if event is None:
            print("Encrypted tally partially decrypted.")
        print("Encrypted tally partially decrypted.")
        return True

    @ex(
        help='Decrypt partially the encrypted tally with threshold of the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
            (['-fprk', '--file-private-key'],
             {'help': 'The file where to input private key of the trustee', 'action': 'store',
              'dest': 'file_private_key'}),
            (['-t', '--trustee-id'],
             {'help': 'The trustee id for the election', 'action': 'store', 'dest': 'trustee_id'}),
            (['-dk', '--decryption-key'],
             {'help': 'The file where to output the decryption key for trustee key decryption', 'action': 'store',
              'dest': 'file_decryption_key'}),
        ],
    )
    def decrypt_threshold(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        if not self.app.pargs.file_private_key:
            self.app.pargs.file_private_key = shell.Prompt(
                "Enter the file where to input private key of the trustee: ").input

        if not self.app.pargs.trustee_id:
            self.app.pargs.trustee_id = shell.Prompt("Enter the trustee id for the election: ").input

        if not self.app.pargs.file_decryption_key:
            self.app.pargs.file_decryption_key = shell.Prompt("Enter file where to output the decryption key: ").input

        try:
            with open(self.app.pargs.file_decryption_key, 'r') as file:
                decryption_key = json.load(file)
        except:
            print("Invalid private key file")
            return False

        owned_partial_decryption = TrusteeCore.partial_decrypt_encrypted_ballot(self.app.pargs.uuid, int(decryption_key),
                                                                                int(self.app.pargs.trustee_id))
        if owned_partial_decryption is None:
            print("Encrypted tally not partially decrypted.")
            return False
        event = ServerCore.add_event_partial_decryption(self.app.pargs.uuid)
        if event is None:
            print("Encrypted tally partially decrypted.")
        print("Encrypted tally partially decrypted.")
        return True

    @ex(
        help='Compute the result of the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
        ],
    )
    def compute_result(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        result = ServerCore.compute_result(self.app.pargs.uuid)
        if result is None:
            print("The result was not computed.")
            return False
        event = ServerCore.set_event_result(self.app.pargs.uuid)
        if event is None:
            print("The result was not computed.")
        print("The result was computed.")
        return True

    @ex(
        help='Show the array of result of the election',

        # sub-command level arguments.
        arguments=[
            (['-u', '--uuid'], {'help': 'The uuid of the election', 'action': 'store', 'dest': 'uuid'}),
        ],
    )
    def show_result(self):
        if not self.app.pargs.uuid:
            self.app.pargs.uuid = shell.Prompt("Enter the uuid of the election: ").input

        result = ServerCore.get_array_of_result(self.app.pargs.uuid)
        if result is None:
            print("The array of result was not computed.")
            return False
        print(result)
        return True



from belenios.custom_types.belenios.StringType import StringType

from sqlalchemy import Column, Integer, ForeignKey

from belenios.models.belenios.SignedObjectModel import SignedObjectModel
from sqlalchemy import Column, Integer, ForeignKey

from belenios.custom_types.belenios.StringType import StringType
from belenios.models.belenios.SignedObjectModel import SignedObjectModel


class SignedStringModel(SignedObjectModel):
    id = Column(Integer, ForeignKey('signed_object_model.id'), primary_key=True)
    value = Column(StringType, nullable=False)
    __mapper_args__ = {
        "polymorphic_identity": "SignedStringModel",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = 'SignedStringModel'

    def __str__(self):
        return (
            f"SignedStringModel(\n"
            f"    value={self.value}\n"
            f")"
        )

    def to_json(self):
        return {
            "value": self.value,
        }

    @staticmethod
    def load_from_json(dict):
        return SignedStringModel(type=dict['type'], value=dict['value'])



from sqlalchemy.orm.exc import NoResultFound

from belenios.custom_types.command_line.RoleType import RoleType
from belenios.models.command_line.UserModel import UserModel
from belenios.utilities.Utility import Utility


class DatabaseInit():
    def run(self, session):
        self._create_users(session)

    def _create_users(self, session):
        a = [
            UserModel(username='admin', password_hash=Utility.hash_sha256_as_hex('admin'.encode()), role=RoleType.ServerAdministrator),
            UserModel(username='cred', password_hash=Utility.hash_sha256_as_hex('credd'.encode()), role=RoleType.CredentialAuthority),
            UserModel(username='trustee1', password_hash=Utility.hash_sha256_as_hex('trustee1'.encode()), role=RoleType.Trustee),
            UserModel(username='trustee2', password_hash=Utility.hash_sha256_as_hex('trustee2'.encode()), role=RoleType.Trustee),
            UserModel(username='trustee3', password_hash=Utility.hash_sha256_as_hex('trustee3'.encode()), role=RoleType.Trustee),
        ]
        for u in a:
            # Check if the example account exists in the database
            try:
                existing_user = session.query(UserModel).filter_by(username=u.username).one()
            except NoResultFound:
                # If the account doesn't exist, create it
                session.add(u)
        session.commit()



from belenios.models.belenios.BallotModel import BallotModel
from belenios.models.belenios.ProofModel import ProofModel
from belenios.models.belenios.SignatureModel import SignatureModel
from belenios.utilities.AnswerUtility import AnswerUtility
from belenios.utilities.ProofOfKnowledge import ProofOfKnowledge
from belenios.utilities.Utility import Utility
from belenios.utilities.command_line.DbSession import DbSession


class VoterCore:
    @staticmethod
    def get_public_credential_from_index(election_uuid, index):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        for credential in election_bundle.public_credentials:
            if credential.index == index:
                return credential

    @staticmethod
    def get_salt_from_private_credential(election_uuid, index):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None
        for salt in election_bundle.salts:
            if salt.index == index:
                return salt

    @staticmethod
    def check_choices_structure(election_uuid, choices):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return False
        # check type of choice ([[int, int,...], [int, int,...], [blank, int,...], ...])
        if not isinstance(choices, list):
            return False
        for c in choices:
            if not isinstance(c, list):
                return False
            for a in c:
                if not isinstance(a, int):
                    return False
        if len(choices) != len(election_bundle.election.questions):
            return False

        # validing the size of sub-arrays and correct blank vote
        for i, choice in enumerate(choices):
            question = election_bundle.election.questions[i]
            if question.type == "QuestionHModel":
                if question.blank:
                    if len(choice) != len(question.answers) + 1:
                        return False
                    if choice[0] == 1:  # it a blank vote
                        for a in choice[1:]:
                            if a != 0:
                                return False
                    else:
                        for a in choice:
                            if a not in [0, 1]:
                                return False
                        if not (question.min <= choice[1:].count(1) <= question.max):
                            return False
                else:
                    if len(choice) != len(question.answers):
                        return False
                    for a in choice:
                        if a not in [0, 1]:
                            return False
                    if not (question.min <= choice.count(1) <= question.max):
                        return False
            else:  # "type": "QuestionNhModel"
                if len(choice) != len(question.answers):
                    return False
                for a in choice:
                    if a not in [0, 1]:
                        return False
        return True

    @staticmethod
    def sign_ballot(election_uuid, ballot, x):
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        # we generate A = g^w using a key pair, where private_key = A and public_key = w
        tmp = Utility.gen_key_el_gamal(election_bundle.election.group)
        w = tmp.private_key
        A = tmp.public_key
        H = ballot.to_hash_sha256()
        challenge = ProofOfKnowledge.Hsignature(H, A) % election_bundle.election.group.q
        response = (w - x * challenge)
        ballot.signature = SignatureModel()
        ballot.signature.hash = H
        ballot.signature.proof = ProofModel()
        ballot.signature.proof.challenge = challenge
        ballot.signature.proof.response = response
        return ballot

    @staticmethod
    def generate_ballot(election_uuid, choices_for_all_questions, private_credential):
        # 1. V gets public data of E
        election_bundle = Utility.get_election_bundle_from_uuid(election_uuid)
        if election_bundle is None:
            print("Election uuid invalid")
            return None

        prefix, index = Utility.get_prefix_index_from_credential(private_credential)

        # 2. V uses the index in her secret credential c to get her public credential ˆc (from election public
        # data) and her salt s (from salts), and checks that ˆc = public(c, s)

        # getting public credential from election bundle
        public_credential = VoterCore.get_public_credential_from_index(election_uuid, index)
        if public_credential is None:
            print("Credential invalid")
        # getting salt from election bundle
        salt = VoterCore.get_salt_from_private_credential(election_uuid, index)
        if salt is None:
            print("Credential invalid")

        # check that public credential ^c is ok
        x = Utility.derive_secret_exponent(prefix, salt, election_bundle.election)
        if str(pow(election_bundle.election.group.g, x,
               election_bundle.election.group.p)) != str(public_credential.public_credential):  # Calculate g^x mod p
            print("Credential invalid")

        ballot = BallotModel()
        ballot.election_uuid = election_bundle.election.uuid
        ballot.election_hash = election_bundle.setup_data.election
        ballot.credential = int(public_credential.public_credential)

        prefix, index = Utility.get_prefix_index_from_credential(private_credential)
        secret = Utility.derive_secret_exponent(prefix, salt, election_bundle.election)
        # Try to make a vote :
        # TODO: check if we need to have the private of public key
        # ballot = AnswerUtility.generate_single_nh_answer(self.app, election_bundle, question, "Non",
        #                                                     UnsafelyStoringCredentials().credentials[0], salts[0])
        for i, question in enumerate(election_bundle.election.questions):
            if question.type == "QuestionHModel":
                answer = AnswerUtility.generate_answer_for_h_question(election_uuid, question.blank,
                                                                      choices_for_all_questions[i], secret)
                if answer is None:
                    return None
                else:
                    ballot.answers.append(answer)
            else:
                answer = AnswerUtility.generate_answer_for_nh_question(election_uuid, choices_for_all_questions[i],
                                                                       private_credential, public_credential, salt)
                if answer is None:
                    return None
                else:
                    ballot.answers.append(answer)
        # TODO: add verification of the ballot and a command to send it separately
        # 4. S processes b:
        # (a) let C be the public credential used in b (its credential field)
        # (b) S checks that (C, wi, V) ∈ L
        # (c) S checks all zero-knowledge proofs of b
        # (d) S adds b to D
        # 5. at any time (even after tally), V may check that h (if it is her last ballot) appears in the list
        # of pretty ballots P B and the weight of her ballot as it appears in P B is equal to her weight
        DbSession().session.add(ballot)
        VoterCore.sign_ballot(election_uuid, ballot, x)

        # We can check is the answer and the proof is correctly generated :
        if AnswerUtility.check_ballot(election_uuid, ballot,secret, choices_for_all_questions):
            print("Proof is valid")
        else:
            print("Proof is invalid")
            return None

        DbSession().session.commit()
        return ballot



from sqlalchemy import Column, Enum

from belenios.custom_types.belenios.StringType import StringType
from belenios.custom_types.command_line.RoleType import RoleType
from belenios.models.BaseModel import BaseModel


class UserModel(BaseModel):
    username = Column(StringType, nullable=False)
    password_hash = Column(StringType, nullable=False)
    role = Column(Enum(RoleType), nullable=False)



import json

from sqlalchemy.types import TypeDecorator, VARCHAR


class ArrayOfStringsType(TypeDecorator):
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            # Serialize the array of strings to JSON
            if not isinstance(value, list):
                value = [value]
            for i,v in enumerate(value):
                value[i] = str(v)
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            # Deserialize the JSON to an array of strings
            value = json.loads(value)
        return value

    def process_literal_param(self, value, dialect):
        return value



from enum import Enum


class RoleType(Enum):
    ServerAdministrator = 'ServerAdministrator'
    CredentialAuthority = 'CredentialAuthority'
    Trustee = 'Trustee'
    Voter = 'Voter'
    VotingServer = 'VotingServer'



from sqlalchemy import Column
from sqlalchemy import Integer, ForeignKey

from belenios.custom_types.belenios.ArrayOfIsType import ArrayOfIsType
from belenios.models.BaseModel import BaseModel


class ResultRowModel(BaseModel):
    result_id = Column(Integer, ForeignKey('result_model.id'))
    result_cells = Column(ArrayOfIsType, nullable=False)

    def __init__(self, *args, **kwargs):
        if 'result_cells' not in kwargs:
            kwargs['result_cells'] = []
        super().__init__(*args, **kwargs)

    def __str__(self):
        return (
            f"ResultRowModel(\n"
            f"    result_id={self.result_id},\n"
            f"    result_cells={self.result_cells},\n"
            f")"
        )


from sqlalchemy import Column

from belenios.custom_types.belenios.ZqType import ZqType
from belenios.models.BaseModel import BaseModel
from belenios.utilities.Utility import Utility


class ProofModel(BaseModel):
    challenge = Column(ZqType, nullable=False)
    response = Column(ZqType, nullable=False)

    def to_hash_sha256(self):
        delimiter = '|'
        data = (
                f"id={self.id}" + delimiter +
                f"challenge={self.challenge}" + delimiter +
                f"response={self.response}"
        )
        sha256_hash = Utility.hash_sha256_as_hex(data.encode())
        return sha256_hash

    def to_json(self):
        return {
            "challenge": self.challenge,
            "response": self.response
        }

    def __str__(self):
        return (
            f"ProofModel(\n"
            f"    challenge={self.challenge},\n"
            f"    response={self.response}\n"
            f")"
        )

    @staticmethod
    def load_from_json(dict):
        return ProofModel(challenge=int(dict['challenge']), response=int(dict['response']))
